from anonlink._entitymatcher import ffi, lib

__author__ = 'Stephen Hardy, Brian Thorne'


def dicecoeff_pure_python(e1, e2):
    """
    Dice coefficient measures the similarity of two bit patterns.

    Implemented exclusively in Python.

    :param e1: bitarray of same length as e2
    :param e2: bitarray of same length as e1
    :return: real 0-1 similarity measure
    """
    count1 = e1.count()
    count2 = e2.count()
    combined_count = count1 + count2
    overlap_count = (e1 & e2).count()
    if combined_count == 0:
        return 0.0
    else:
        return 2.0 * overlap_count / combined_count


def dicecoeff_native(e1, e2):
    """
    Dice coefficient measures the similarity of two bit patterns.

    Implemented via an external library.

    :param e1: bitarray of same length as e2
    :param e2: bitarray of same length as e1
    :return: real 0-1 similarity measure
    """
    e1array = ffi.new("char[]", e1.tobytes())
    e2array = ffi.new("char[]", e2.tobytes())
    return lib.dice_coeff(e1array, e2array, len(e1array))


def dicecoeff(e1, e2):
    """
    Dice coefficient measures the similarity of two bit patterns

    :return: real 0-1 similarity measure
    """
    if e1.length() == e2.length() and (e1.length()/8) % 8 == 0:
        return dicecoeff_native(e1, e2)
    else:
        return dicecoeff_pure_python(e1, e2)


def dicecoeff_precount(e1, e2, count):
    """
    Dice coefficient measures the similarity of two bit patterns

    :param e1: bitarray1
    :param e2: bitarray2
    :param count: float bitcount1 + bitcount2
    :return: real 0-1 similarity measure
    """
    if count == 0:
        return 0
    return 2*(e1 & e2).count()/count


def tanimoto(e1, e2):
    """
    Tanimoto coefficient measures the similarity of two bit patterns.

    Also referred to as the Jaccard similarity

    :return: real 0-1 similarity measure
    """
    return (e1 & e2).count() / float((e1 | e2).count())


def tanimoto_precount(e1, e2, count):
    """
    Tanimoto coefficient measures the similarity of two bit patterns

    :param e1: bitarray1
    :param e2: bitarray2
    :param count: float bitcount1 + bitcount2
    :return: real 0-1 similarity measure
    """
    a = (e1 & e2).count()
    return a / float(count - a)



import os.path

from cffi import FFI

ffibuilder = FFI()

_path = os.path.dirname(__file__)
sourcefile = os.path.join(_path, 'dice_one_against_many.cpp')

with open(sourcefile, 'r') as f:
    source = f.read()


ffibuilder.set_source(
    "_entitymatcher",
    source,
    source_extension='.cpp',
    extra_compile_args=['-Wall', '-Wextra', '-Werror', '-O3', '-std=c++11', '-march=native', '-mssse3', '-mpopcnt', '-fvisibility=hidden'
    ],
)

ffibuilder.cdef("""
    int match_one_against_many_dice_k_top(const char *one, const char *many, const uint32_t *counts_many, int n, int keybytes, uint32_t k, double threshold, int *indices, double *scores);
    double dice_coeff(const char *array1, const char *array2, int array_bytes);
    double popcount_arrays(uint32_t *counts, const char *arrays, int narrays, int array_bytes);
""")


if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



import random
from timeit import default_timer as timer

from clkhash.key_derivation import generate_key_lists
from clkhash.schema import get_schema_types
from clkhash.bloomfilter import calculate_bloom_filters
from clkhash.randomnames import NameList

from anonlink.entitymatch import *
from anonlink.util import popcount_vector, generate_clks, generate_bitarray


some_filters = generate_clks(10000)


def compute_popcount_speed(n):
    """
    Just do as much counting of bits.
    """
    clk_bits = 1024
    clk_bytes = clk_bits / 8
    clks_MiB = n * clk_bytes * 1.0 / (1 << 20)

    clks = [generate_bitarray(clk_bits) for _ in range(n)]

    print("{:6d} x {:d} bit popcounts".format(n, clk_bits))
    print("Implementation              | Time (ms) | Bandwidth (MiB/s) | Throughput (1e6 popc/s)")

    # Python
    popcounts, elapsed_time = popcount_vector(clks, use_python=True)

    # Guard against division by zero (have observed elapsed_nocopy ==
    # 0 in the wild; see issue #92).
    inv_elapsed_time = 1 / elapsed_time if elapsed_time > 0 else float('inf')

    python_speed_in_MiB = clks_MiB * inv_elapsed_time
    python_Mops = (n / 1e6) * inv_elapsed_time

    elapsed_time_ms = elapsed_time * 1e3
    print("Python (bitarray.count()):  |  {:7.2f}  |   {:9.2f}       | {:7.2f}"
          .format(elapsed_time_ms, python_speed_in_MiB, python_Mops))

    # Native
    start = timer()
    popcounts, elapsed_nocopy = popcount_vector(clks, use_python=False)
    end = timer()
    elapsed_time = end - start

    # Guard against division by zero (have observed elapsed_nocopy ==
    # 0 in the wild; see issue #92).
    inv_elapsed_time = 1 / elapsed_time if elapsed_time > 0 else float('inf')
    inv_elapsed_nocopy = 1 / elapsed_nocopy if elapsed_nocopy > 0 else float('inf')

    copy_percent = 100*(elapsed_time - elapsed_nocopy) * inv_elapsed_time
    elapsed_time_ms = elapsed_time * 1e3
    elapsed_nocopy_ms = elapsed_nocopy * 1e3
    native_speed_in_MiB = clks_MiB * inv_elapsed_time
    native_speed_in_MiB_nocopy = clks_MiB * inv_elapsed_nocopy
    native_Mops = (n / 1e6) * inv_elapsed_time
    native_Mops_nocopy = (n / 1e6) * inv_elapsed_nocopy
    print("Native code (no copy):      |  {:7.2f}  |   {:9.2f}       | {:7.2f}"
          .format(elapsed_nocopy_ms, native_speed_in_MiB_nocopy, native_Mops_nocopy))
    print("Native code (w/ copy):      |  {:7.2f}  |   {:9.2f}       | {:7.2f} ({:.1f}% copying)"
          .format(elapsed_time_ms, native_speed_in_MiB, native_Mops, copy_percent))

    return python_speed_in_MiB


def print_comparison_header(threshold):
    print("\nThreshold:", threshold)
    print("Size 1 | Size 2 | Comparisons      | Total Time (s)          | Throughput")
    print("       |        |        (match %) | (comparisons / matching)|  (1e6 cmp/s)")
    print("-------+--------+------------------+-------------------------+-------------")


def compute_comparison_speed(n1, n2, threshold):
    """
    Using the greedy solver, how fast can hashes be computed using one core.
    """

    assert n1 != 0 and n2 != 0
    filters1 = [some_filters[random.randrange(0, 8000)] for _ in range(n1)]
    filters2 = [some_filters[random.randrange(2000, 10000)] for _ in range(n2)]

    start = timer()
    sparse_matrix = calculate_filter_similarity(filters1, filters2, len(filters2), threshold)
    t1 = timer()
    _ = greedy_solver(sparse_matrix)
    end = timer()

    similarity_time = t1 - start
    solver_time = end - t1
    elapsed_time = end - start

    inv_similarity_time = 1 / similarity_time if similarity_time > 0 else float('inf')
    inv_elapsed_time = 1 / elapsed_time if elapsed_time > 0 else float('inf')

    print("{:6d} | {:6d} | {:4d}e6  ({:5.2f}%) | {:6.3f}  ({:4.1f}% / {:4.1f}%) |  {:8.3f}".format(
        n1, n2, n1*n2 // 1000000,
        100.0*len(sparse_matrix)/(n1*n2),
        elapsed_time,
        100.0*similarity_time * inv_elapsed_time,
        100.0*solver_time * inv_elapsed_time,
        (n1*n2 / 1e6) * inv_similarity_time))
    return elapsed_time


def compare_python_c(ntotal=10000, nsubset=6000, frac=0.8):
    """Compare results and running time of python and C++ versions.

    :param ntotal: Total number of data points to generate
    :param nsubset: Number of points for each database
    :param frac: Fraction of overlap between subsets

    :raises: AssertionError if the results differ
    :return: dict with 'c' and 'python' keys with values of the total time taken
             for each implementation
    """

    nml = NameList(ntotal)
    sl1, sl2 = nml.generate_subsets(nsubset, frac)

    keys = generate_key_lists(('test1', 'test2'), len(nml.schema))
    filters1 = calculate_bloom_filters(sl1, get_schema_types(nml.schema), keys)
    filters2 = calculate_bloom_filters(sl2, get_schema_types(nml.schema), keys)

    # Pure Python version
    start = timer()
    result = python_filter_similarity(filters1, filters2, 1, 0.0)
    end = timer()
    python_time = end - start

    # C++ cffi version
    start = timer()
    result3 = cffi_filter_similarity_k(filters1, filters2, 1, 0.0)
    end = timer()
    cffi_time = end - start

    assert result == result3, "Results are different between C++ cffi and Python"

    # Results are the same
    return {
        "c": cffi_time,
        "python": python_time
    }


def benchmark(size, compare):

    print("Anonlink benchmark -- see README for explanation")
    print("------------------------------------------------")
    if compare:
        print(compare_python_c(ntotal=1000, nsubset=600))

    compute_popcount_speed(100000)

    possible_test_sizes = [
        1000, 2000, 3000, 4000,
        5000, 6000, 7000, 8000, 9000,
        10000,
        20000, 30000, 40000, 50000, 60000, 70000, 80000, 90000,
        1000000,
        2000000
    ]

    thld = 0.5
    print_comparison_header(thld)
    for test_size in possible_test_sizes:
        if test_size <= size:
            compute_comparison_speed(test_size, test_size, thld)

    thld = 0.7
    print_comparison_header(thld)
    size *= 5
    for test_size in possible_test_sizes:
        if test_size <= size:
            compute_comparison_speed(test_size, test_size, thld)


if __name__ == '__main__':
    benchmark(4000, False)
    # benchmark(20000, False)



#!/usr/bin/env python3.4

"""
Concurrent implementations.
"""

import concurrent.futures
import anonlink.entitymatch
from anonlink.util import chunks


def calc_chunk_result(chunk_number, chunk, filters2, k, threshold):
    chunk_results = anonlink.entitymatch.calculate_filter_similarity(chunk, filters2, k, threshold)

    partial_sparse_result = []
    # offset chunk's A index by chunk_size * chunk_number
    chunk_size = len(chunk)
    offset = chunk_size * chunk_number
    for (ia, score, ib) in chunk_results:
        partial_sparse_result.append((ia + offset, score, ib))

    return partial_sparse_result


def calculate_filter_similarity(filters1, filters2, k, threshold):
    """
    Example way of computing similarity scores in parallel.

    :param filters1:
    :param filters2:
    :param k:
    :param threshold:
    :return:
    """

    results = []
    chunk_size = int(20000000 / len(filters2))

    with concurrent.futures.ProcessPoolExecutor() as executor:
        futures = []

        for i, chunk in enumerate(chunks(filters1, chunk_size)):
            future = executor.submit(calc_chunk_result, i, chunk, filters2, k, threshold)
            futures.append(future)

        for future in futures:
            results.extend(future.result())

    return results



"""
Solves pairwise matches for the entire network at once.

Given a sparse, weighted, bipartite graph determine the matching.
"""

import logging
import networkx as nx
from networkx.algorithms import bipartite

log = logging.getLogger('anonlink.networkflow')


def calculate_network(similarity, cutoff):
    """Given an adjacency matrix of edge weights, apply a
    threshold to the connections and construct a graph.

    :param similarity: The list of tuples including n-gram similarity scores
    :param cutoff: The threshold for including a connection
    :return: The resulting networkx graph.
    """
    G = nx.DiGraph()
    logging.debug('Applying threshold to network')
    for (idx1, score, idx2) in similarity:
        if score > cutoff:
            G.add_edge('row'+str(idx1), 'col'+str(idx2), weight=score, capacity=1.0)

    return G


def _to_int_map(network, find_pair):
    """Given a dictionary of edges {'rowN': 'colM', ...}, and a
    function to find the mate node.
    Return a dictionary of {N: M, ...}
    """
    entityMap = {}
    for node in network:
        if node.startswith("row") and len(network[node]):
            paired_node = find_pair(network, node)
            if paired_node is not None:
                entityMap[int(node[3:])] = int(paired_node[3:])
    return entityMap


def calculate_entity_mapping(G, method=None):
    """Given the networkx graph, calculate a dictionary mapping
    each row node to the most highly similar column node.

    :param G: A `networkx.Graph` comprising of nodes from two entities,
    connected by equally weighted edges if the similarity was above a
    threshold.

    :param method: The method to use to solve the entity mapping.
    Options are
        - 'flow' or None (default) - `networkx.maximum_flow` method.
        - 'bipartite' - the `networkx.bipartite.maximum_matching` algorithm (fastest)
        - 'weighted' - the `networkx.max_weight_matching` (slowest but most accurate
          with close matches)

    :return: A dictionary mapping of row index to column index. If no mate
    is found, the node isn't included.
    """

    if method == 'bipartite':
        log.info('Solving entity matches with bipartite maximum matching solver')
        network = bipartite.maximum_matching(G)
        entity_map = _to_int_map(network, lambda network, node: network[node])

    elif method == 'weighted':
        log.info('Solving entity matches with networkx maximum weight matching solver')
        network = nx.max_weight_matching(G)
        entity_map = _to_int_map(network, lambda network, node: network[node])

    elif method == 'flow' or method is None:
        log.info('Solving entity matches with networkx maximum flow solver')
        # The maximum flow solver requires a SOURCE and SINK
        num_rows, num_cols = 0, 0
        for i, node in enumerate(G.nodes()):
            if node.startswith("row"):
                G.add_edge('start', node, capacity=1.0)
                num_rows += 1
            if node.startswith("col"):
                G.add_edge(node, 'end', capacity=1.0)
                num_cols += 1
        flow_value, network = nx.maximum_flow(G, 'start', 'end')

        # This method produces a quality metric `flow`, however
        # it needs to be compared to the number of entities
        if flow_value < num_rows:
            log.info('Matching not perfect - {:.3f}'.format(flow_value))
        else:
            log.info('Matching complete. (perfect matching)')

        def find_pair(network, node):
            # Make sure to deal with unconnected nodes

            possible_nodes = [n for n in network[node] if n != 'start']
            if len(possible_nodes) > 0:
                def get_score(node_name):
                    return network[node][node_name]
                return max(possible_nodes, key=get_score)
            else:
                return None

        entity_map = _to_int_map(network, find_pair)

    else:
        raise NotImplementedError("Haven't implemented that matching method")

    return entity_map


def map_entities(weights, threshold, method=None):
    """Calculate a dictionary mapping using similarity scores.

    :param weights: The list of tuples including n-gram similarity scores
    :param threshold: The threshold for including a connection
    :param method: The method to use to solve the entity mapping.
    Options are
        - 'flow' or None (default for datasets under 100K records) - `networkx.maximum_flow` method.
        - 'bipartite' - the `networkx.bipartite.maximum_matching` algorithm
        - 'weighted' - the `networkx.max_weight_matching` (slowest but most accurate
          with close matches)
    """

    network = calculate_network(weights, threshold)
    return calculate_entity_mapping(network, method)


if __name__ == "__main__":
    A = [[4.0, 3.0, 2.0, 1.0],
         [1.0, 4.0, 3.0, 2.0],
         [2.0, 1.0, 4.0, 3.0],
         [2.5, 3.5, 4.5, 1.5]]

    print("Threshold | Match | Entity Mapping")
    n_thresholds = 10
    for t in range(n_thresholds + 1):
        # threshold will range from 2.5 to 3.5 inclusive in steps of 0.1
        threshold = 2.5 + 1.0 * t / n_thresholds
        entity_map = map_entities(A, threshold)
        perfect_match = len(entity_map) == len(A)
        print("{:9.3f} | {:5} | {:26s} ".format(threshold, perfect_match, entity_map))



import pkg_resources

from anonlink import bloommatcher
from anonlink import entitymatch
from anonlink import network_flow

__version__ = pkg_resources.get_distribution('anonlink').version
__author__ = 'N1 Analytics'



from itertools import repeat
import logging

from anonlink._entitymatcher import ffi, lib

import sys
from operator import itemgetter

from . import bloommatcher as bm

log = logging.getLogger('anonlink.entitymatch')


def python_filter_similarity(filters1, filters2, k, threshold):
    """Pure python method for determining Bloom Filter similarity

    Both arguments are 3-tuples - bitarray with bloom filter for record, index of record, bitcount

    :return: A list of tuples *k* for each entity in filters1.
    The tuple comprises:
        - the index in filters1
        - the similarity score between 0 and 1 of the k matches above threshold
        - The index in filters2 of the best match
    """
    result = []
    for i, f1 in enumerate(filters1):
        def dicecoeff(x):
            return bm.dicecoeff_precount(f1[0], x[0], float(f1[2] + x[2]))

        coeffs = filter(lambda c: c[1] >= threshold,
                        enumerate(map(dicecoeff, filters2)))
        top_k = sorted(coeffs, key=itemgetter(1), reverse=True)[:k]
        result.extend([(i, coeff, j) for j, coeff in top_k])
    return result


def cffi_filter_similarity_k(filters1, filters2, k, threshold):
    """Accelerated method for determining Bloom Filter similarity.

    Assumes all filters are the same length, being a multiple of 64
    bits.

    """
    length_f1 = len(filters1)
    length_f2 = len(filters2)

    if length_f1 == 0:
        return []

    # There's no sense in having k > length_f2. Also, k is passed to
    # ffi.new(...) below, so we need to protect against an
    # out-of-memory DoS if k is of untrustworthy origin.
    k = min(k, length_f2)

    filter_bits = len(filters1[0][0])
    assert filter_bits % 64 == 0, 'Filter length must be a multple of 64 bits.'
    filter_bytes = filter_bits // 8

    match_one_against_many_dice_k_top = lib.match_one_against_many_dice_k_top

    # An array of the *one* filter
    clist1 = [ffi.new("char[{}]".format(filter_bytes), bytes(f[0].tobytes()))
              for f in filters1]

    if sys.version_info < (3, 0):
        # Python 2 version
        data = []
        for f in filters2:
            for b in f[0].tobytes():
                data.append(b)

        carr2 = ffi.new("char[{}]".format(filter_bytes * length_f2), ''.join(data))
    else:
        # Works in Python 3+
        carr2 = ffi.new("char[{}]".format(filter_bytes * length_f2),
                        bytes([b for f in filters2 for b in f[0].tobytes()]))

    c_popcounts = ffi.new("uint32_t[{}]".format(length_f2), [f[2] for f in filters2])

    # easier to do all buffer allocations in Python and pass them to C,
    # even for output-only arguments
    c_scores = ffi.new("double[]", k)
    c_indices = ffi.new("int[]", k)

    result = []
    for i, f1 in enumerate(filters1):
        assert len(clist1[i]) == filter_bytes
        matches = match_one_against_many_dice_k_top(
            clist1[i],
            carr2,
            c_popcounts,
            length_f2,
            filter_bytes,
            k,
            threshold,
            c_indices,
            c_scores)

        if matches < 0:
            raise ValueError('Internal error: Bad key length')

        # Take the first `matches` elements of c_scores and c_indices.
        # Store them along with `i`.
        result.extend(zip(repeat(i, matches), c_scores, c_indices))

    return result


def greedy_solver(sparse_similarity_matrix):
    """
    For optimal results consider sorting input by score for each row.

    :param sparse_similarity_matrix: Iterable of tuples: (indx_a, similarity_score, indx_b)
    """
    mapping = {}

    # Indices of filters which have been claimed
    matched = set()

    for i, score, j in sparse_similarity_matrix:
        if i not in mapping and j not in matched:
            mapping[i] = j
            matched.add(j)

    return mapping


def calculate_mapping_greedy(filters1, filters2, k, threshold):
    """
    Brute-force one-shot solver.

    :param filters1: A list of cryptoBloomFilters from first organization
    :param filters2: A list of cryptoBloomFilters from second organization
    :param float threshold: The score above which to consider
    :param int k: Consider up to the top k matches.

    :returns A mapping dictionary of index in filters1 to index in filters2.
    """

    log.info('Solving with greedy solver')

    sparse_matrix = calculate_filter_similarity(filters1, filters2, k, threshold)
    return greedy_solver(sparse_matrix)


def calculate_filter_similarity(filters1, filters2, k, threshold, use_python=False):
    """Computes a sparse similarity matrix with:
        - size no larger than k * len(filters1)
        - order of len(filters1) + len(filters2)

    This method is used as part of the all in one function `calculate_mapping_greedy`
    however you may wish to call it yourself in order to break a large mapping into
    more sizable chunks. These partial similarity results can be reduced into one
    list and passed to the `greedy_solver` function. Note the returned index for
    filters1 will need to be offset, see anonlink.concurrent for an example.

    :param filters1: A list of cryptoBloomFilters from first organization
    :param filters2: A list of cryptoBloomFilters from second organization
    :param k: The top k edges will be included in the result.
    :param threshold: Only scores greater than this threshold will be considered
        (between 0 and 1)
    :param use_python: Use the slower pure python method instead of C++ implementation

    :return: A sparse matrix as a list of tuples, up to *k* for each entity
        in filters1. Will be shorter if there are no qualifying matches.

        The tuple comprises:
            - the index in filters1
            - the similarity score between 0 and 1 of the best match
            - The index in filters2 of the best match
    """
    if not filters1 or not filters2:
        raise ValueError('empty input')
    # use C++ version by default
    if use_python:
        return python_filter_similarity(filters1, filters2, k, threshold)
    else:
        return cffi_filter_similarity_k(filters1, filters2, k, threshold)



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))


requirements = [
        "bitarray==0.8.1",
        "networkx==1.11",
        "cffi>=1.7",
        "clkhash>=0.10"
    ]

setup(
    name="anonlink",
    version='0.8.1',
    description='Anonymous linkage using cryptographic hashes and bloom filters',
    url='https://github.com/n1analytics/anonlink',
    license='Apache',
    setup_requires=['cffi>=1.7'],
    install_requires=requirements,
    packages=find_packages(exclude=[
        '_cffi_build', '_cffi_build/*',
        'tests'
    ]),
    package_data={'anonlink': ['data/*.csv', '_cffi_build']},

    ext_package="anonlink",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Security :: Cryptography",
    ],

    # for cffi
    cffi_modules=["_cffi_build/build_matcher.py:ffibuilder"],
    zip_safe=False,
)



#!/usr/bin/env python3.4

import os
import random
from bitarray import bitarray
from timeit import default_timer as timer

from anonlink._entitymatcher import ffi, lib


def generate_bitarray(length):
    a = bitarray(endian=['little', 'big'][random.randint(0, 1)])
    a.frombytes(os.urandom(length//8))
    return a


def generate_clks(n):
    res = []
    for i in range(n):
        ba = generate_bitarray(1024)
        res.append((ba, i, ba.count()))
    return res


def popcount_vector(bitarrays, use_python=True):
    """Return a list containing the popcounts of the elements of
    bitarrays, and the time (in seconds) it took. If use_python is
    False, use the native code implementation instead of Python; in
    this case the returned time is the time spent in the native code,
    NOT including copying to and from the Python runtime.

    Note, due to the overhead of converting bitarrays into bytes,
    it is currently more expensive to call our C implementation
    than just calling bitarray.count()
    """
    # Use Python
    if use_python:
        start = timer()
        counts = [clk.count() for clk in bitarrays]
        elapsed = timer() - start
        return counts, elapsed

    # Use native code
    n = len(bitarrays)
    arr_bytes = bitarrays[0].length() // 8
    c_popcounts = ffi.new("uint32_t[{}]".format(n))
    many = ffi.new("char[{}]".format(arr_bytes * n),
                    bytes([b for f in bitarrays for b in f.tobytes()]))
    ms = lib.popcount_arrays(c_popcounts, many, n, arr_bytes)

    return [c_popcounts[i] for i in range(n)], ms * 1e-3


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]
